#ifndef BisqPAPUinternals_hh
#define BisqPAPUinternals_hh

#include "clock.h"

#define PAPU_LENGTH_DEBUG 0     /* stop timer (all chans) */
#define PAPU_DECAY_DEBUG 0      /* volume, decay          */
#define PAPU_SQUARE_DEBUG 0     /* square, duty, sweep    */
#define PAPU_TRIANGLE_DEBUG 0
#define PAPU_RAND_DEBUG 0
#define PAPU_DMC_DEBUG 0
#define PAPU_WAVE_DEBUG 0

class waveunit
{
public:
    WORD wavepos;
    WORD wavelen;
public:
    waveunit() : wavepos(0), wavelen(400)
    {
    }
    inline void write(fBYTE ind, fBYTE value)
    {
    	if(ind == 2)
    		wavelen = (wavelen & 0xFF00) | value;
    	else if(ind == 3)
    		wavelen = BYTE(wavelen) | ((value&7) << 8);
#if PAPU_WAVE_DEBUG
		if(ind == 2 || ind == 3)
			fprintf(stderr, "(wavelen=%u)", wavelen);
#endif
    }
    inline WORD getwavelen() const { return wavelen; }
    inline bool setwavelen(fWORD len)
    {
        register bool carry;
        if(len > 0x7FF)
        {
            len &= 0x7FF;
            carry = true;
        }
        else
            carry = false;
        
        wavelen = len;
        return carry;
    }
};

// This requires CPU tick
#define BEGIN_WAVETICKER void tick() { if(!wavepos) {
#define END_WAVETICKER wavepos = wavelen; } else --wavepos; }
#define BEGIN_HZTICKER(x) inline void hz##x##tick() {
#define END_HZTICKER }

class lengthunit
{
    fBYTE lengthcounter;
    fBYTE bmask;
    bool lengthdisabled;
public:
    bool chanenabled;
public:
    lengthunit(fBYTE mask = 5) : lengthcounter(0), bmask(1<<mask), lengthdisabled(true), chanenabled(false)
    {
    }
    inline void write(fBYTE ind, fBYTE value)
    {
        if(ind == 3)
        {
            static const BYTE lengthtrans[32]=
            {5,127,10,1,20, 2,40, 3,80, 4,30, 5,7, 6,14, 7,
             6,  8,12,9,24,10,48,11,96,12,36,13,8,14,16,15};
            // Ks papu.txt jos tuon taulukon sislt kiinnostaa
            
            if(chanenabled)
            {
                lengthcounter = lengthtrans[(value >> 3) & 31];
                #if PAPU_LENGTH_DEBUG
                fprintf(stderr, "(\33[1;33mlcounter=%u\33[m)", lengthcounter);
                #endif
            }
            else
            {
                #if PAPU_LENGTH_DEBUG
            	fprintf(stderr, "(\33[1;31mlcounter NOT MODIFIED\33[m)");
            	#endif
            }
        }
        if(ind == 0)
        {
            lengthdisabled = (value & bmask);
            #if PAPU_LENGTH_DEBUG
            if(lengthdisabled)
            	fprintf(stderr, "(\33[33mlength DISabled at %u\33[m)", lengthcounter);
            else
            	fprintf(stderr, "(\33[33mlength ENabled at %u\33[m)", lengthcounter);
            #endif
        }
    }
    inline void enable() { chanenabled = true; }
    inline void disable(){ chanenabled = false;}
    inline bool lengthunit_silent() const
    {
        return !chanenabled || !lengthcounter;
    }
    BEGIN_HZTICKER(60)
        if(!lengthdisabled && lengthcounter > 0)
            --lengthcounter;
    END_HZTICKER
    void debug_silence()
    {
    	fprintf(stderr, ";length(%sabled, counter=%u)", lengthdisabled?"dis":"en", lengthcounter);
    }
};

class decayunit
{
    fBYTE decaycounter;
    fBYTE decayrate;
    fBYTE decaypos;
    fBYTE decaymode;
public:
    decayunit() : decaycounter(0), decayrate(15), decaypos(0), decaymode(0x10)
    {
    }
    inline void write(fBYTE ind, fBYTE value)
    {
        if(ind == 0)
        {
            decayrate = value & 15;
            decaymode = (value>>4) & 3;
            #if PAPU_DECAY_DEBUG
            if(decaymode & 1)
            	fprintf(stderr, "[\33[0;1;34mvolume=%u\33[m]", decayrate);
            else if(decaymode & 2)
            	fprintf(stderr, "[\33[0;36mdrate=%u,looping\33[m]", decayrate);
            else
            	fprintf(stderr, "[\33[0;36mdrate=%u\33[m]", decayrate);
        	#endif
        }
        else if(ind == 3)
        {
            #if PAPU_DECAY_DEBUG
        	fprintf(stderr, "[\33[0;1;34mdcounter=15,was %u\33[m]", decaycounter);
        	#endif
        	
        	/* Only a write out to $4003/$4007/$400F will reset the
             * current envelope decay counter to a known state (to
             * $F, the maximum volume level) for the appropriate
             * channel's envelope decay hardware.
             */
        	decaycounter = 15;
        }
        // Decay rate is the speed it changes with.
    }
    inline unsigned getvol() const
    {
    	if(decaymode & 1)
    	{
    		/* When the envelope decay bit (bit 4) is set(1), the
    		 * current volume value (bits 0-3) is sent directly to
    		 * the channel DAC.
    		 */
    		return decayrate; 
    	}
    	
    	/* When the envelope decay disable bit is cleared, bits 0-3
    	 * now control the envelope decay rate, and an internal 4-bit
    	 * down counter (hereon the envelope decay counter) now
    	 * controls the channel's volume level.
    	 */    	
        return decaycounter;
    }
    BEGIN_HZTICKER(240)
        if(!decaypos)
        {
            // This counter is for timing; it always runs.
            decaypos = decayrate;                
            
            if(decaycounter > 0)
                --decaycounter;
            else if(decaymode & 2)
                decaycounter = 15;
        }
        else
            --decaypos;
    END_HZTICKER

    void debug_silence()
    {
        if(decaymode & 1)
        	fprintf(stderr, ";volume=%u", decayrate);
        else
        {
        	fprintf(stderr, ";decay[%u,rate=%u", decaycounter, decayrate);
        	if(decaymode & 2)
	        	fprintf(stderr, ",looping");
	    }
    }
};

class squaregenunit : public decayunit, public waveunit, public lengthunit
{
    fBYTE sweepdata;
    fBYTE sweeppos;
    fBYTE sweepspeed;
    fBYTE sweepshift;
    bool sweepenabled;
    bool jammed;
    fBYTE decr;
public:
    fBYTE dutytype;
    fBYTE dutycounter;
public:
    squaregenunit() : sweeppos(0), sweepspeed(0), sweepenabled(false), jammed(false), decr(0), dutytype(0)
    { 
    }
    inline void set_decr_mode(fBYTE d) { decr = d; }
    inline void write(fBYTE ind, fBYTE value)
    {
    	#if PAPU_SQUARE_DEBUG
        fprintf(stderr, "square.write(%u, 0x%02X) - ", ind, value);
        #endif
        waveunit::write(ind, value);
        decayunit::write(ind, value);
        lengthunit::write(ind, value);
        
        if(ind == 1)
        {
            sweepdata = value;
            sweepspeed = ((value >> 4) & 7) + 1;
            sweepshift = value & 7;
            sweepenabled = (value & 0x87) > 0x80;
            #if PAPU_SQUARE_DEBUG
            fprintf(stderr,
              "{sweepshift %d %s, rate %d, %sabled}",
              sweepshift, (value&8)?"down":"up", (value>>4)&7, (value&128)?"en":"dis");
            #endif
        }
        else if(ind == 0)
        {
        	#if PAPU_SQUARE_DEBUG
            if((value >> 6) != dutytype)
                fprintf(stderr, "{duty=%d)\n", value >> 6);
            #endif
            dutytype = value >> 6;
        }
        jammed = getwavelen() < 8;
        #if PAPU_SQUARE_DEBUG
        if(jammed)
        	fprintf(stderr, " -> jammed\n");
        else if(silent())
        	fprintf(stderr, " -> silent\n");
        else
        	fprintf(stderr, " -> ok\n");
    	#endif
    }
    BEGIN_HZTICKER(120)
        if(!sweeppos)
        {
            // Update sweep speed counter
            sweeppos = sweepspeed;
            if(sweepenabled /* && !silent() */) // May not update if silent!
            {
                unsigned len = getwavelen();
                if(sweepdata & 15)
                    len -= (len >> sweepshift) + decr;
                else
                    len += (len >> sweepshift);
                if(setwavelen(len) || len < 8)
                    jammed = true;
            }
        }
        else
            --sweeppos;
    END_HZTICKER
    inline bool silent() const
    {
        return jammed || lengthunit_silent();
    }
    BEGIN_WAVETICKER
        /*if(!silent())*/ ++dutycounter;
    END_WAVETICKER
    
    void debug_silence()
    {
    	fprintf(stderr, "%sabled", chanenabled?"en":"dis");
    	fprintf(stderr, "; wlen=%u, sweep(shift %d %s, rate %d, %sabled)",
    		wavelen,
    		sweepshift, (sweepdata&8)?"down":"up", (sweepdata&7), sweepenabled?"en":"dis");
    	decayunit::debug_silence();
	    lengthunit::debug_silence();
    }
};

class triangleunit : public waveunit, public lengthunit
{
    enum {triload, tricount} trimode;
    fBYTE linearcounter, triloadvalue;
    bool oldbit;
    bool jammed;
public:
    fBYTE trianglecounter;
public:
    triangleunit() : lengthunit(7), trimode(triload), linearcounter(0),
                     triloadvalue(0), oldbit(false), jammed(false), trianglecounter(0)
    {
    }
    inline void write(fBYTE ind, fBYTE value)
    {
    	#if PAPU_TRIANGLE_DEBUG
        fprintf(stderr, "triangle.write(%u, 0x%02X) -", ind, value);
        #endif
        waveunit::write(ind, value);
        lengthunit::write(ind, value);
        
        if(ind == 3)
        {
            trimode = (oldbit || !linearcounter) ? triload : tricount;            
        }
        else if(ind == 0)
        {
            triloadvalue = value & 0x7F;
            bool newbit = value & 0x80;
            
            if(!oldbit)
            {
                trimode = tricount;
                linearcounter = triloadvalue;
            }
            else if(!newbit && !linearcounter)
                trimode = tricount;
            
            newbit = oldbit;
        }
        jammed = !linearcounter || getwavelen() < 8;
        #if PAPU_TRIANGLE_DEBUG
        if(jammed)
	        fprintf(stderr, "-> jammed\n");
	    else if(lengthunit_silent())
	    	fprintf(stderr, "-> silent\n");
	    else
	       	fprintf(stderr, "-> ok\n");
        #endif
    }
    BEGIN_HZTICKER(120)
    END_HZTICKER
    BEGIN_HZTICKER(240)
        if(trimode==tricount && linearcounter > 0)
        {
            --linearcounter;
            if(!linearcounter)jammed = true;
        }
    END_HZTICKER
    inline bool silent() const
    {
        return jammed || lengthunit_silent();
    }
    
    BEGIN_WAVETICKER
        /*if(!silent())*/ ++trianglecounter;
    END_WAVETICKER

    void debug_silence()
    {
    	fprintf(stderr, "%sabled", chanenabled?"en":"dis");
    	fprintf(stderr, "; wlen=%u", wavelen);
    	lengthunit::debug_silence();
    }
};

class randgenunit : public decayunit, public lengthunit
{
    WORD wavelen;
    bool randmode;  // false = 32767-bit, true = 93-bit
    WORD shiftreg;
    unsigned wavepos;
public:
    fBYTE output;
public:
    randgenunit() : wavelen(0), shiftreg(0x0001), wavepos(0), output(0)
    {
    }
    inline void write(fBYTE ind, fBYTE value)
    {
    	#if PAPU_RAND_DEBUG
        fprintf(stderr, "rand.write(%u, 0x%02X) -", ind, value);
        #endif
        decayunit::write(ind, value);
        lengthunit::write(ind, value);
        
        if(ind == 2)
        {
            static const WORD wlenconvdata[16] =
            {2*0x002, 2*0x004, 2*0x008, 2*0x010,
             2*0x020, 2*0x030, 2*0x040, 2*0x050,
             2*0x065, 2*0x07F, 2*0x0BE, 2*0x0FE,
             2*0x17D, 2*0x1FC, 2*0x3F9, 2*0x7F2};
            wavelen = wlenconvdata[value & 15];
            randmode = value >> 7;
            #if PAPU_RAND_DEBUG
            fprintf(stderr, "wavelen=%u, mode=%u", wavelen, randmode);
            #endif
        }
        #if PAPU_RAND_DEBUG
        if(silent())
        	fprintf(stderr, "-> silent\n");
        else
        	fprintf(stderr, "-> ok\n");
        #endif
    }
    BEGIN_HZTICKER(120)
    END_HZTICKER
    inline bool silent() const
    {
        return lengthunit_silent();
    }
    BEGIN_WAVETICKER
        if(!silent())
        {
            output = ~((shiftreg >> 14) ^ (shiftreg >> (randmode ? 8 : 13))) & 1;
            shiftreg = (shiftreg << 1) | output;
        }
    END_WAVETICKER
    
    void debug_silence()
    {
    	fprintf(stderr, "%sabled", chanenabled?"en":"dis");
    	fprintf(stderr, ";shiftreg=");
    	for(unsigned a=15; a>0; )fprintf(stderr, "%d", (shiftreg>>--a)&1);
    	fprintf(stderr, ",result=%d,wlen=%u,bits=%d",
    		output, wavelen, randmode?93:32767);
    	decayunit::debug_silence();
	    lengthunit::debug_silence();
    }
};

template<class T>
class papuhzticker : public T
{
private:
    unsigned countermax, hzcounter;
    fBYTE hzpos;
public:
    papuhzticker() : hzpos(0)
    {
        countermax = 1364 * Clock->ScanlineCount / 4;
        // This makes 240 ticks per second
        hzcounter = countermax;
    }
    void tick()
    {
        T::tick();
        if(!--hzcounter)
        {
            hzcounter = countermax;
            switch(hzpos)
            {
            	case 0:
            		hz60tick();
            		hz120tick();
            		hz240tick();
            		hzpos=1;
            		break;
            	case 1:
            		hz240tick();
            		hzpos=2;
            		break;
            	case 2:
            		hz120tick();
            		hz240tick();
            		hzpos=3;
            		break;
            	case 3:
            		hz240tick();
            		hzpos=0;
            		break;
            }
        }
    }
};

typedef papuhzticker<squaregenunit> squaregen;
class sq1gen : public squaregen { public: sq1gen() { set_decr_mode(0); } };
class sq2gen : public squaregen { public: sq2gen() { set_decr_mode(1); } };

class dmcgen
{
	BYTE playbackmode;
    fWORD wlen, wcount;       // waveform position counter
    fWORD playaddr, loadaddr; // playing address
    fWORD dmalen, lengthleft; // playing length
public:
	bool irqflag;
    bool chanenabled;
    BYTE deltacounter;
public:
    dmcgen() : playbackmode(0), wcount(0xFFF),
               playaddr(0xC000), loadaddr(0xC000), lengthleft(0),
               irqflag(false), chanenabled(false), deltacounter(0)
    {
    }
    void write(fBYTE ind, fBYTE value)
    {
    	#if PAPU_DMC_DEBUG
        fprintf(stderr, "dmc::write(%u, 0x%02X)", ind, value);
        #endif
    	if(!ind)
    	{
            static const WORD wlenconvdata[16] =
            {0xD60, 0xBE0, 0xAA0, 0xA00,
             0x8F0, 0x7F0, 0x710, 0x6B0,
             0x5F0, 0x500, 0x470, 0x400,
             0x350, 0x2A0, 0x240, 0x1B0};
             
            wlen = wlenconvdata[value & 15];
			playbackmode = value >> 6;
			irqflag = false;
			#if PAPU_DMC_DEBUG
			fprintf(stderr, "(wlen %u, mode %d, irq flag reset)", wlen, playbackmode);
			#endif
    	}
    	else if(ind==1)
    	{
    		deltacounter = value & 127;
    		#if PAPU_DMC_DEBUG
    		fprintf(stderr, "(counter=0x%02X)", value & 127);
    		#endif
    	}
    	else if(ind==2)
    	{
    		loadaddr = ((fWORD)value << 6) | 0xC000;
    		playaddr = loadaddr;
    		#if PAPU_DMC_DEBUG
    		fprintf(stderr, "(loading address=%X)", loadaddr);
    		#endif
    	}
    	else if(ind==3)
    	{
    		dmalen = (fWORD)value << 4;
    		lengthleft = dmalen;
    		#if PAPU_DMC_DEBUG
    		fprintf(stderr, "(length=%u)", dmalen);
    		#endif
    	}
    	#if PAPU_DMC_DEBUG
    	fprintf(stderr, "\n");
    	#endif
    }
    inline void enable()
    {
        chanenabled = true;
        irqflag = false;
    }
    inline void disable()
    {
        chanenabled = false;
    }
    inline bool silent() const
    {
    	return !chanenabled || !lengthleft;
    }
    bool tick()
    {
    	if(silent())return false;
    	
	    //  This requires CPU tick
	    if(!wcount)
	    {
			if(!wlen)return false;
			wcount = wlen;
		}
		else
		{
	    	--wcount;
	    	return false;
	    }
    	
    	BYTE byt = Op6502(playaddr);
    	if(playaddr == 0xFFFF)playaddr = 0x8000;else ++playaddr;
    	
    	#if PAPU_DMC_DEBUG
    	fprintf(stderr, "%02X", byt);
    	#endif
    	
    	for(unsigned a=0; a<8; ++a)
    	{
    		if(byt & 1)
    		{
    			if(deltacounter < 126)deltacounter += 2;
    		}
    		else if(deltacounter > 1)deltacounter -= 2;
	    	byt >>= 1;
	    	//byt <<= 1;
	    }
	    #if PAPU_DMC_DEBUG
	    fprintf(stderr, "(%u)", deltacounter);
	    #endif
    	
    	if(!--lengthleft)
    	{
    		if(playbackmode==2)
    		{
    			#if PAPU_DMC_DEBUG
    			fprintf(stderr, "DMC IRQ\n");
    			#endif
    			irqflag = true;
    			return true;
    		}
    		if(playbackmode & 1)
    		{
    			lengthleft = dmalen;
    			playaddr = loadaddr;
    			#if PAPU_DMC_DEBUG
    			fprintf(stderr, "DMC loop\n");
    			#endif
    		}
    		else
    		{
    			#if PAPU_DMC_DEBUG
    			fprintf(stderr, "DMC stop\n");
    			#endif
    		}
   		}
    	return false;
    }
    void debug_silence()
    {
    	fprintf(stderr, "%sabled", chanenabled?"en":"dis");
    	fprintf(stderr, "; wlen=%u, mode=%d, counter=0x%02X, loading address=%X, length=%u, irqflag=%s",
    		wlen, playbackmode, deltacounter, loadaddr, dmalen, irqflag?"set":"no");
    }
};

/*

Yhteenveto:

 Vasemmampana oleva tarvitsee AINA sisempn (alempana/oikeampana) olevan
 palveluita, ja waveunit tarvitsee aina emonsa wavetick()-funktiota.
 
    |-sq1gen
    |    `- hzticker
    |        `- squaregenunit (generaattoriyksikk)
    |            |- waveunit
    |            `- decayunit
    |                `- lengthunit
    |-sq2gen
    |     `- hzticker
    |        `- squaregenunit (generaattoriyksikk)
    |            |- waveunit
    |            `- decayunit
    |                `- lengthunit
    |-trigen
    |    `- hzticker
    |        `- trianglegenunit (generaattoriyksikk)
    |            |- waveunit
    |            `- lengthunit
    |-rndgen
    |    `- hzticker
    |        `- randgenunit (generaattoriyksikk)
    |            `- decayunit
    |                `- lengthunit
    `-dmcgen
        `- sndchan

waveunit pitisi siis kirjoittaa ilman virtualia.
Miten kaksisuuntainen perint pitisi kirjoittaa?

*/

#endif
